*********************************************************************
*                                                                   *
*                           This is PARNI                           *
*             (Practical Adaptive Random Number Idealizer)          *
*                                                                   *
* author: Andre van Hameren                                         *
*   date: Sat Apr 19 2003                                           *
*********************************************************************
*
*********************************************************************
* usage:
*      call initparni(ndim,readchan)
*
*      do iev=1,nev
*         
*        generate_ndim_random_numbers(rho,ndim)
*        generate_one_random_number(rch)
*         
*        call gnrtparni(xx,rho,rch)
*
*        evaluate_your_integrand(ff,xx)
*
*        call wghtparni(wght,xx)
*
*        call adapparni(ff*wght,2d0)
*
*        apply_your_statistics(ff*wght)
*
*      enddo
*
* initparni, input:
*     ndim = the number of dimensions of the hypercube you want to 
*            generate random vectors in
* readchan = a logical, default is .false.. 
*            If you put readchan=.true. then parni will load channels 
*            from a file "channels.parni" and use them as starting 
*            point. The subroutine "writparni" is available to print 
*            the channels to this file.
* 
* gnrtparni, input:
*      rho = array of ndim (random) numbers between 0 and 1
*      rch = (random) numbers between 0 and 1
*           output:
*       xx = array of ndim (random) numbers between 0 and 1
*
* wghtparni, input: 
*       xx = array of ndim (random) numbers between 0 and 1
*           output:
*     wght = wght coming with xx ( wght=1/density(xx) )
*           
* adapparni, input:
*  ff*wght = total weight, so integrand times "wght" from parni
*    expon = 2d0. Exponent in adaptive multi-channeling, should be
*            put to 1d0 for the creation of a unitary probability
*            decomposition.
*
* For now, I suggest the user to play around with the parameters in 
* the subroutine "initparni" below in order to get an optimal result.
*********************************************************************

       subroutine initparni(nndI,readchan)
*  **************************************************************
*  **************************************************************
      implicit none
      double precision arr(5555)
      integer nndI ,iid,ich ,iter
      logical readchan
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      double precision ave(5555)
      common/aveparni/ ave
      integer          ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
      common/datparni/ ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      integer          nnd
      common/nndparni/ nnd
* put input parameters
      nnd = nndI ! dimension of hypercube
* set further parameters
      nchmax  = 1000 ! maximum number of channels (=boxes)
      ndatmin = 1000 ! batch-size of data for weight optimization
      niter   = 40   ! # boxes that is splitted in one optimization step
      nmerge  = 20   ! # channels that is discard in one optimization step
      nstp    = 1    ! # weight optimization steps before channel optimization
      write(6,101) nchmax,ndatmin,niter,nmerge,nstp 
  101 format('##### PARNI #####'
     &      ,' nch=',i4,', ndat=',i5,', niter=',i3
     &      ,', nmerge=',i3,', nstp=',i2,' #') 
* initialize lch, the labels of the channels
      do ich=1,nchmax
        lch(ich) = ich
      enddo
* initialize statistics
      istp = 0
      ndat = 0
      do ich=1,nchmax
        ave(ich) = 0d0
      enddo
* initialize channels
      if (readchan) then
        call readparni
      else
        nch = 1
        wch(nch) = 1d0
        do iid=1,nnd
          bxp(nch,iid) = 1d0
          bxm(nch,iid) = 0d0
        enddo
        vol(nch) = 1d0
* * For starting with more than 2*nnd channels use iter=1,1+2*nnd  etc
        do iter=1,1 +2*nnd +(2*nnd)**2
* * * order channels: wch(lch(1)) is going to be largest
          do ich=1,nch
            arr(ich) = -wch(lch(ich))
          enddo
          call sortparni(arr,lch,nch)
* * * create new channels by dissecting the largest
          call dissectfirst
        enddo
      endif
      end
*
      subroutine adapparni(wght,expon)
*  **************************************************************
*  **************************************************************
      implicit none
      double precision wght,expon ,factor,sum,arr(5555)
      integer          ich,iid,nless,nmore ,iter
      logical          adaptwg,adaptch 
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision ave(5555)
      common/aveparni/ ave
      integer          ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
      common/datparni/ ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      double precision sumdns,nchloc,loc(5555)
      common/dnsparni/ sumdns,nchloc,loc
ccostas
      integer index
* update ave,  loc  labels channels with non-zero density-value
      factor = wght**expon/sumdns
      do ich=1,nchloc
c       ave(loc(ich)) = ave(loc(ich)) + factor/vol(loc(ich))
ccostas
        index=loc(ich)
        ave(index) = ave(index) + factor/vol(index)
      enddo
      ndat = ndat + 1
* check if enough data have been collected
      adaptwg = (ndat.eq.ndatmin)
* adapt channel-weights
      if (adaptwg) then
        istp = istp+1
        sum = 0d0
        do ich=1,nch
          ave(lch(ich)) = ave(lch(ich))/dble(ndat)
          wch(lch(ich)) = wch(lch(ich))*ave(lch(ich))**(1d0/expon)
          ave(lch(ich)) = 0d0
          sum = sum + wch(lch(ich))
        enddo
* * normalize wch
        do ich=1,nch
          wch(lch(ich)) = wch(lch(ich))/sum 
        enddo  
* * re-initialize ndat
        ndat = 0
      endif
* adapt channels
      adaptch = (istp.eq.nstp)
      if (adaptch) then
        istp = 0
* * order channels: wch(lch(1)) is going to be largest
        do ich=1,nch
          arr(ich) = -wch(lch(ich))
        enddo
        call sortparni(arr,lch,nch)
* * merge smallest  nmerge  channels 
        if (nch.gt.nmerge.and.nmerge.gt.1) call mergelastfew(nmerge-1)
* * create new channels by dissecting the largest
        call dissectfirst
        do iter=1,niter-1
* * * order channels: wch(lch(1)) is going to be largest
          do ich=1,nch
            arr(ich) = -wch(lch(ich))
          enddo
          call sortparni(arr,lch,nch)
* * * create new channels by dissecting the largest
          call dissectfirst
        enddo
      endif
      end
*
      subroutine mergelastfew(nless)
*  **************************************************************
*  * merge  nless+1  smallest channels to 1 channel
*  **************************************************************
      implicit none
      double precision         wch1,bxp1(37),bxm1(37)
      integer          nless ,ich,iid
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
* initialize result of merging
      wch1 = 0d0
      do iid=1,nnd
        bxp1(iid) = 0d0
        bxm1(iid) = 1d0
      enddo
* merge
      do ich=nch-nless,nch 
        wch1 = wch1 + wch(lch(ich))
        do iid=1,nnd
          if (bxp(lch(ich),iid).gt.bxp1(iid)) 
     &      bxp1(iid) = bxp(lch(ich),iid)
          if (bxm(lch(ich),iid).lt.bxm1(iid)) 
     &      bxm1(iid) = bxm(lch(ich),iid)
        enddo
      enddo
* put result of merging
      ich = lch(nch-nless)
      wch(ich) = wch1
      vol(ich) = 1d0
      do iid=1,nnd
        bxp(ich,iid) = bxp1(iid)
        bxm(ich,iid) = bxm1(iid)
        vol(ich) = vol(ich)*( bxp(ich,iid)-bxm(ich,iid) )
      enddo
* update nch
      nch = nch-nless
      end
*
      subroutine dissectfirst
*  **************************************************************
*  * Replace channel lch(1) by its 2*nnd pieces after dissection 
*  **************************************************************
      implicit none
      double precision         wch1,bxp1(37),bxm1(37),vol1
      integer          nmore ,ich,jch,iid,nmany,kch(74)
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      integer          ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
      common/datparni/ ndat,ndatmin,nchmax,niter,nmerge,istp,nstp
* number of extra channels
      nmore = 2*nnd-1
* merge smallest  nmany+1  channels if necessary
      nmany = nch+nmore-nchmax
      if (nmany.gt.0) call mergelastfew(nmany)
* store the first channel
      wch1 = wch(lch(1))/dble(2*nnd)
      do iid=1,nnd
        bxp1(iid) = bxp(lch(1),iid)
        bxm1(iid) = bxm(lch(1),iid)
      enddo
      vol1 = vol(lch(1))/2d0
* store last  nmore  entries of  lch
      do ich=1,nmore
        kch(ich) = lch(nch+ich)
      enddo
* shift lch with  nmore  entries
      do ich=nch,1,-1
        lch(ich+nmore) = lch(ich)
      enddo
      nch = nch+nmore
* re-fill first  nmore  entries of  lch
      do ich=1,nmore
        lch(ich) = kch(ich)
      enddo
* copy the previously first channel  2*nnd  times
      do ich=1,2*nnd
        wch(lch(ich)) = wch1
        do iid=1,nnd
          bxp(lch(ich),iid) = bxp1(iid)
          bxm(lch(ich),iid) = bxm1(iid)
        enddo
        vol(lch(ich)) = vol1
      enddo
* cut the copies in two, each one along another direction
      do iid=1,nnd
        ich = 2*iid
        bxp(lch(ich),iid) = ( bxp(lch(ich),iid) 
     &                       +bxm(lch(ich),iid) )/2d0
        ich = ich-1
        bxm(lch(ich),iid) = ( bxp(lch(ich),iid)
     &                       +bxm(lch(ich),iid) )/2d0
      enddo
      end
*
      subroutine gnrtparni(xx,rho,rch)
*  **************************************************************
*  * "rch" is used to choose a channel (box), 
*  * "rho" is used to construct the point "xx" in the chosen box 
*  **************************************************************
      implicit none
      double precision xx(37) ,rho(37),rch,wchsum,xm,xp
      integer          ich,iid
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
* choose channel
      wchsum = 0d0
      do ich=1,nch
        wchsum = wchsum + wch(lch(ich))
        if (rch.lt.wchsum) goto 100
      enddo
*
      write(6,*) 'Error in gnrtparni: no channel chosen.'
      write(6,*) 'Channels dumped in file "dumped.parni"'
      open(99,file='dumped.parni')
   51 format('channel',18x,'weight',13x,'sum weights')
   52 format(i7,2d24.16)
      write(99,51)
      xm = 0d0
      do ich=1,nch
        xm = xm + wch(lch(ich))
        write(99,52) lch(ich),wch(lch(ich)),xm
      enddo
      close(99)
      stop
*      
  100 continue
* generate  xx  in box  ich 
      do iid=1,nnd
        xm = bxm(lch(ich),iid)
        xp = bxp(lch(ich),iid)
        xx(iid) = xm + (xp-xm)*rho(iid)
      enddo
      end
*
      subroutine wghtparni(wght,xx)
*  **************************************************************   
*  **************************************************************
      implicit none
      double precision wght,xx(37)
      integer          ich,iid,jch
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      double precision sumdns,nchloc,loc(5555)
      common/dnsparni/ sumdns,nchloc,loc
      logical          debug
      common/debugcom/ debug
      sumdns = 0d0
      jch = 0
      do ich=1,nch
        do iid=1,nnd
          if (           xx(iid).le.bxm(lch(ich),iid)
     &   .or.  bxp(lch(ich),iid).lt.          xx(iid) ) goto 100
        enddo  
        jch = jch+1
        loc(jch) = lch(ich)
        sumdns = sumdns + wch(lch(ich))/vol(lch(ich))
  100   continue
      enddo
      nchloc = jch
      wght = 1d0/sumdns
      end

 
      subroutine sortparni(arr,lbl,nn)
*  **************************************************************
*  * heapsort:
*  * order the array "arr" of double precisions from small to 
*  * large, and re-order the array "lbl" of integers with it.
*  * So after 
*  *      do ii=1,nn
*  *        arr(ii) = brr(lbl(ii))
*  *      enddo
*  *      call sortparni(arr,lbl,nn)
*  * brr(lbl(ii)) will be ordered from small to large, and after
*  *      do ii=1,nn
*  *        arr(ii) = -brr(lbl(ii))
*  *      enddo
*  *      call sortparni(arr,lbl,nn)
*  * brr(lbl(ii)) will be ordered from large to small.
*  **************************************************************
      implicit none 
      double precision  arr(5555)    ,arrb 
      integer           lbl(5555),nn ,lblb,ll,ii,jj,ir
      if (nn.le.1) goto 30
*     
      ll = nn/2 + 1
      ir = nn
  10  continue
        if (ll.gt.1) then 
          ll = ll - 1
          arrb = arr(ll)
          lblb = lbl(ll)
        else
          arrb = arr(ir)
          lblb = lbl(ir)
          arr(ir) = arr(1)
          lbl(ir) = lbl(1)
          ir = ir - 1
          if (ir.eq.1) then
            arr(1) = arrb
            lbl(1) = lblb
            goto 30
          endif
        endif
        ii = ll
        jj = ll + ll
  20    continue
        if (jj.le.ir) then
          if (jj.lt.ir) then
            if (arr(jj).lt.arr(jj+1)) jj = jj + 1
          endif
          if (arrb.lt.arr(jj)) then
            arr(ii) = arr(jj)
            lbl(ii) = lbl(jj)
            ii = jj
            jj = jj + jj
          else
            jj = ir + 1
          endif
        goto 20
        endif
        arr(ii) = arrb
        lbl(ii) = lblb
      goto 10
  30  continue
*
      end
      
      
      subroutine writparni
*  **************************************************************   
*  **************************************************************
      implicit none
      integer ich,iid
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      open(99,file='channels.parni')
      write(99,101) nch,nnd
      do ich=1,nch
        write(99,102) lch(ich),wch(lch(ich))
     &,(bxm(lch(ich),iid),iid=1,nnd),(bxp(lch(ich),iid),iid=1,nnd)
      enddo
      close(99)
  101 format(i4,i3)
  102 format(i4,75d24.16)
      end
*
      subroutine readparni
*  **************************************************************   
*  **************************************************************
      implicit none
      integer ich,iid
      double precision bxm(5555,37),bxp(5555,37),vol(5555)
      common/boxparni/ bxm,bxp,vol
      integer          nnd
      common/nndparni/ nnd
      double precision wch(5555)
      common/wchparni/ wch
      integer          nch,lch(5555)
      common/nchparni/ nch,lch
      open(99,file='channels.parni')
      read(99,101) nch,nnd
      do ich=1,nch
        read(99,102) lch(ich),wch(lch(ich))
     &,(bxm(lch(ich),iid),iid=1,nnd),(bxp(lch(ich),iid),iid=1,nnd)
      enddo
      close(99)
  101 format(i4,i3)
  102 format(i4,75d24.16)
      end
